// Copyright 2014 Google Inc. All Rights Reserved.
// Licensed under the Apache License, Version 2.0, found in the LICENSE file.

part of doc_coverage_frontend;

const methodTypes = const [
  'getters',
  'setters',
  'constructors',
  'methods',
  'operators'
];

class LibraryDocAnalyzer {
  final String name, nameWithColon, base, package;
  final Element section = new Element.section();
  final Element sectionHeading = new HeadingElement.h1();
  final ImageElement sectionShield = new ImageElement();
  final TableElement scoresTable = new TableElement();
  final ParagraphElement gapSummary = new ParagraphElement();
  List<Element> sortedSections = new List();
  String htmlUrl;

  LibraryDocAnalyzer(name, this.base, {this.package: null})
      : name = name,
        nameWithColon = name.replaceFirst('-', ':');

  void prepareElements({shield: true}) {
    sortedSections.clear();

    sectionHeading.text = 'library $name';
    section.append(sectionHeading);

    if (shield) {
      sectionShield
        ..attributes['src'] = DocCoverage.shieldUrlForScore(0)
        ..classes.add('shield');
      sectionHeading.append(sectionShield);
    }

    gapsDiv.append(section);
    gapSummary.dataset['value'] = '0';
    section.append(gapSummary);
  }

  void analyzeScore() {
    prepareElements();
    String gapsUrl = (package == null)
        ? '#/library/${name.replaceFirst('-', ':')}/gaps'
        : '#/package/${package}/gaps';

    AnchorElement linkToGaps = new AnchorElement()
      ..attributes['href'] = gapsUrl
      ..text = 'See';
    ParagraphElement p = new ParagraphElement()
      ..append(linkToGaps)
      ..appendText(' the gaps in doc coverage.');
    section.append(p);
    section.append(scoresTable);
    scoresTable.classes
      ..add('hidden')
      ..add('scores-table');
    getJsonAndReport('score');
  }

  void updateLibraryBadge() {
    List<Map> memberData = sortedSections
        .map((Element el) => {
      'size': int.parse(el.dataset['size']),
      'score': int.parse(el.dataset['count']) / 100
    })
        .toList();
    int score = (100 * DocCoverage.weightedScore(memberData)).toInt();
    sectionShield.attributes['src'] = DocCoverage.shieldUrlForScore(score);
  }

  void analyzeGaps() {
    prepareElements(shield: false);
    section.classes.add('hidden');
    getJsonAndReport('gaps');
  }

  /// Fetches the Json form of the library's docs, and analyzes the comments for:
  ///
  /// * the library itself
  /// * the library's "classes" (each class, error, and typedef)
  /// * the library's top-level functions
  /// * the library's top-level properties
  void getJsonAndReport(String screen) {
    HttpRequest.getString('$base/${name}.json').then((String json) {
      Map<String, dynamic> library = new JsonDecoder().convert(json);
      new LibraryCommentDocAnalyzer(this, library).go(screen);
      (library['classes']['class'] as List).forEach(
          (klass) => new ClassDocAnalyzer(this, 'class', klass).go(screen));
      (library['classes']['error'] as List).forEach(
          (klass) => new ClassDocAnalyzer(this, 'error', klass).go(screen));
      // Typedefs don't have their own files; it's all inline.
      (library['classes']['typedef'] as Map).values
          .forEach((klass) => new TypedefDocAnalyzer(this, klass).go(screen));

      methodTypes.forEach(
          (methodType) => (library['functions'][methodType] as Map)
              .forEach((String name, Map method) {
        new FunctionDocAnalyzer(this, methodType, method).go(screen);
      }));

      (library['variables'] as Map).forEach((String name, Map variable) {
        new VariableDocAnalyzer(this, variable).go(screen);
      });
    }).catchError(_handleError);
  }

  void _handleError(Error error) {
    if (error is! ProgressEvent) throw error;

    ProgressEvent err = error as ProgressEvent;
    handleError(err, section,
        name: 'Dart library "${name.replaceFirst('dart-', '')}"');
  }

  void addToSortedRows(Element classRow, int gapCount, {bool reverse: false}) {
    addToSortedList(classRow, scoresTable, sortedSections, gapCount,
        reverse: reverse);
  }

  void addToSortedSections(Element classSection, int gapCount,
      {bool reverse: false}) {
    addToSortedList(classSection, section, sortedSections, gapCount,
        reverse: reverse);
  }

  void addToSortedList(
      Element element, Element listElement, List<Element> list, int value,
      {bool reverse: false}) {
    // This is craziness. There has to be a better way.
    listElement.classes.remove('hidden');
    if (list.isEmpty) {
      list.add(element);
      listElement.append(element);
      return;
    }

    int i = 0;

    bool keepGoing() {
      int count = int.parse(list[i].dataset['count']);
      return reverse ? value > count : value < count;
    }

    while (i < list.length && keepGoing()) {
      i++;
    }

    if (i == list.length) {
      listElement.append(element);
    } else {
      listElement.insertBefore(element, list[i]);
    }
    list.insert(i, element);
  }

  void bumpCount(int gapCount) {
    int sum = (int.parse(gapSummary.dataset['value'])) + gapCount;
    gapSummary.dataset['value'] = sum.toString();
    gapSummary.innerHtml = '<em>Coverage gap total: $sum points</em>';
  }
}

void reportOnTopLevelComment(Map<String, dynamic> gaps, [Element section]) {
  if (section == null) {
    section = gapsDiv;
  }
  if (!gaps.containsKey('comment') || (gaps['comment'] as String).isEmpty) {
    section.append(new ParagraphElement()
      ..append(dartlangAnchor(gaps['qualifiedName']))
      ..appendText(' has no comment!'));
  } else if ((gaps['comment'] as String).split('\n').length < 2) {
    section.append(new ParagraphElement()
      ..append(dartlangAnchor(gaps['qualifiedName']))
      ..appendText('''\'s comment is too short (under 2 paragraphs)
                        (${DocCoverage.classCommentBrief} points each):'''));
    section.append(new ParagraphElement()
      ..innerHtml = gaps['comment']
      ..className = 'quote');
  }
}

void reportOnMethods(Map<String, dynamic> gaps, [Element section]) {
  bool any = false;
  methodTypes.forEach((cat) {
    if (gaps[cat].length > 0) {
      any = true;
    }
  });
  if (!any) {
    return;
  }

  methodTypes.forEach((cat) => reportOnCategory(cat, gaps, section));
}

void reportOnVariables(Map<String, dynamic> gaps, [Element section]) {
  if (!gaps.containsKey('variables')) return;
  bool any = false;
  if (gaps['variables'].length > 0) {
    any = true;
  }
  if (!any) {
    return;
  }

  reportOnCategory('variables', gaps, section);
}

void reportOnCategory(String cat, Map<String, dynamic> gaps,
    [Element section]) {
  if (section == null) {
    section = gapsDiv;
  }
  List missing = gaps[cat]['missing'];
  List noOneLiner = gaps[cat]['no-one-liner'];

  if (missing.length > 0) {
    String catMsg = missing.length == 1 ? '${singularize(cat)} is' : '$cat are';
    section.append(new ParagraphElement()
      ..text = '''${missing.length} $catMsg missing comments
                    (${DocCoverage.memberCommentGap} points each):''');
    UListElement l = new UListElement();
    section.append(l);
    missing.forEach((Map<String, Object> meth) {
      String name = meth['name'] as String;
      if (name.isEmpty) {
        name = "(default constructor)";
      }
      l.append(new LIElement()
        ..append(dartlangAnchor(meth['qualifiedName'] as String, name)));
    });
  }

  if (noOneLiner.length > 0) {
    String catMsg =
        noOneLiner.length == 1 ? '${singularize(cat)} has' : '$cat have';
    section.append(new ParagraphElement()
      ..text =
      '''${noOneLiner.length} $catMsg no one-liner (the first line is too
                    long) (${DocCoverage.memberCommentIssue} point each):''');
    UListElement l = new UListElement();
    section.append(l);
    noOneLiner.forEach((Map<String, Object> meth) {
      String name = meth['name'] as String;
      if (name.isEmpty) {
        name = "(default constructor)";
      }
      l.append(new LIElement()
        ..append(dartlangAnchor(meth['qualifiedName'] as String, name))
        ..append(new ParagraphElement()
          ..innerHtml = meth['comment']
          ..className = 'quote'));
    });
  }
}
